# coding: utf-8

"""
Module for managing sen2chain processing jobs
"""

import logging
from pathlib import Path
import pandas as pd
import datetime, time
from itertools import chain, groupby
import re
import setuptools
from crontab import CronTab
from collections import OrderedDict
from configparser import ConfigParser
import copy
import numpy as np


from .config import SHARED_DATA, Config
from .indices import IndicesCollection
from .library import Library
from .tiles import Tile
from .utils import datetime_to_str
from .multi_processing import (
    l2a_multiprocessing,
    idx_multiprocessing,
    cld_version_probability_iterations_reprocessing_multiprocessing,
)
from .download_eodag import (
    S2cEodag_download, 
    S2cEodag_download_uniq,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Jobs:
    """
    Class to list and remove existing jobs.
    """

    def __init__(self):
        self._jobs_all = list(Config()._JOBS_DIR.glob("job_*.cfg"))
        self._jobs_all.extend(list(Config()._JOBS_DIR.glob("job_*.py")))
        self._jid_set = set([t.stem[4:] for t in self._jobs_all])
        self._config_files_exist = []
        self._py_scripts_exist = []
        self._logs = []
        self._timing = []
        self._cron_status = []
        self._cron_timing = []
        for jid in self._jid_set:
            logger.disabled = True
            j = Job(jid)
            self._config_files_exist.append(j._config_path.exists())
            self._py_scripts_exist.append(j._python_script_path.exists())
            self._logs.append(j.logs)
            self._timing.append(j.timing)
            self._cron_status.append(j.cron_status)
            self._cron_timing.append(j.cron_timing)
            logger.disabled = False

    def __repr__(self):
        if self._jobs_all:
            return repr(
                pd.DataFrame(
                    OrderedDict(
                        [
                            ("job_id", list(self._jid_set)),
                            ("config_file", self._config_files_exist),
                            ("python_script", self._py_scripts_exist),
                            ("logging", self._logs),
                            ("timing", self._timing),
                            ("cron_status", self._cron_status),
                            ("cron_timing", self._cron_timing),
                        ]
                    )
                )
            )
        else:
            logger.info("No Jobs here")
            return repr(None)

        # >>> job.render().split(' ')[0:4]

    @property
    def list(self):
        return [t.stem[4:] for t in self._jobs_all]

    def remove(self, jid: str = None):
        """
        Remove Job from Job list and cron. Delete job configuration file and execution file.

        :param jid: Job identifier.
        :type jid: str
        """
        if jid in self.list:
            logger.disabled = True
            j = Job(jid)
            logger.disabled = False
            j.cron_remove()
            if j._python_script_path.exists():
                j._python_script_path.unlink()
                logger.info("Removing Python script...")
            if j._config_path.exists():
                j._config_path.unlink()
                logger.info("Removing config file...")
        else:
            logger.info("Job {} not found, doing nothing...".format(jid))


class Job:
    """
    Class to create, edit, save and launch job routines. A job allows you to download L1C, process L2A, cloudmasks and radiometric indices at once, for 1 or more Tile.
    Jobs can be set in cron, and a logging option is available.

    A job is made of :
    - a configuration file, located in sen2chain_data/config/jobs/
    - a .py file if job is set in cron. This .py launches the run() command

    Job config file contains global job parameters and tasks.
    A task only targets a single Tile, meaning if you want to download and process multiple tile products you need to have the equal number of task lines in your configuration file.
    We recommend manually editing the configuration file if you plan to set up a job with multiple Tiles, as editing with the command line is only for 1 task at a time.

    """

    # logger.propagate = False
    def __init__(self, jid: str):
        self._config_path = Config()._JOBS_DIR / ("job_" + jid + ".cfg")
        self._python_script_path = Config()._JOBS_DIR / ("job_" + jid + ".py")
        self.jid = jid
        if self._config_path.exists():
            logger.info("Reading existing config...")
            self.read(self._config_path)
        else:
            logger.info("Creating new job...")
            self.init()
        self._cron = CronTab(user=True)
        self.cron_status = self.get_cron_status()

        self._log_folder_path = Path(Config().get("log_path")) / ("job_" + jid)
        self.remove_used_l1c = False


    def __repr__(self):
        return (
            "logs: {}"
            "\ntiming: {}"
            "\ncron_status: {}"
            "\nprovider: {}"
            "\ndownload tries: {} x / sleep: {} min"
            "\nnb_proc: {}"
            "\ncopy_l2a_sideproducts: {}"
            "\nclean_before: {}"
            "\nclean_after: {}"
            "\nprocess_by_line: {}"            
            "\n\ntasks:\n{}".format(
                self.logs, 
                self.timing,
                self.cron_status,
                self.provider,
                self.tries,
                self.sleep,
                self.nb_proc,
                self.copy_l2a_sideproducts,
                self.clean_before,
                self.clean_after,
                self.process_by_line,                
                self.tasks,
            )
        )
        
    def init(self):
        raw_job_cfg = SHARED_DATA.get("raw_job_cfg")
        self.read(raw_job_cfg)
        first_row = OrderedDict(
            [
                ("tile", ["40KCB"]),
                ("date_min", [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')]),
                ("date_max", [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')]),
                ("max_clouds", [100]),
                ("l1c", ["False"]),
                ("l2a", ["False"]),
                ("cloudmasks", ["False"]),
                ("indices", ["False"]),
                ("remove", ["False"]),
                ("comments", [""]),
            ]
        )
        self.tasks = pd.DataFrame(first_row)

    def task_add(self, row: dict = None):
        """
        Add a task to the job.

        :param row: Dict of specific parameters for task actions. If no dict provided, will provide a default task format.
        :type row: Dict
        """
        if not row:
            logger.info("No row provided, using default")
            row = pd.DataFrame(
                {
                    "tile": ["40KCB"],
                    "date_min": [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')],
                    "date_max": [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')],
                    "max_clouds": [100],
                    "l1c": ["False"],
                    "l2a": ["False"],
                    "cloudmasks": ["False"],
                    "indices": ["False"],
                    "remove": ["False"],
                    "comments": [""],
                }
            )
        # self.tasks = self.tasks.append(row, ignore_index=True)
        self.tasks = pd.concat([self.tasks, row], ignore_index=True)[
            self.tasks.columns
        ]
        self.format_cm_indices()
        logger.info("\n{}".format(self.tasks))

    def task_edit(self, task_id: int = None, **kwargs):
        """
        Edit an existing task, identified by task_id (row number)

        :param task_id: Task identifier, or row number in config file.
        :type task_id: int
        :param kwargs: Any keys/items pairs from task actions options.

        """
        if task_id is None:
            logger.info(
                "Please provide task_id to edit, if no task in job, create_task first"
            )
        else:
            if task_id in self.tasks.index:
                for arg, val in kwargs.items():
                    if arg in self.tasks.columns:
                        self.tasks.loc[task_id, arg] = val
                        logger.info(
                            "Line {}: {} updated to {}".format(
                                task_id, arg, val
                            )
                        )
                    else:
                        logger.info("{} not found".format(arg))
                self.format_cm_indices()
                logger.info("\n{}".format(self.tasks))
            else:
                logger.info("task_id not found")

    def task_remove(self, task_id: int = None):
        """
        Remove a task from job. The task line is deleted from the config file.

        :param task_id: Task identifier (row number).
        :type task_id: int

        """
        if task_id is None:
            logger.info("Please provide task_id to remove")
        else:
            if task_id in self.tasks.index:
                self.tasks.drop(labels=task_id, axis=0, inplace=True)
                self.tasks.reset_index(inplace=True, drop=True)
                logger.info("\n{}".format(self.tasks))
            else:
                logger.info("task_id not found")

    def save(self):
        """
        Save Job config.cfg file to disk.
        """
        with open(str(self._config_path), "w") as ict:
            comments_header ="\n".join(
                [
                    "### Parameters", 
                    "# logs: True | False", 
                    "# timing: in cron format", 
                    "# provider: the provider to download l1c prodicts, default peps, values: peps | scihub", 
                    "# tries: the number of times the download should loop before stopping, to download OFFLINE products", 
                    "# sleep: the time in min to wait between loops", 
                    "# nb_proc: the number of cpu cores to use for this job, default 8", 
                    "# copy_l2a_side_products: to duplicate msk_cldprb_20m and scl_20m from l2a folder to cloudmask folder after l2a production.", 
                    "#   Interesting if you plan to remove l2a to save disk space, but want to keep these 2 files for cloudmask generation and better extraction", 
                    "#   Possible values: True | False", 
                    "# clean_before: set to False or True to call a Libray.clean(remove=True) on selected tiles", 
                    "#   before starting the job, default True", 
                    "# clean_before: set to False or True call a Libray.clean(remove=True) on selected tiles", 
                    "#   after finishing the job, default True", 
                    "# process_by_line: split job execution into n lines for processing of large datasets ", 
                    "#", 
                    "### Table of task entries", 
                    '# Each record line starting with "!" can be commented (and skipped)',
                    "# tile: tile identifier, format ##XXX, comment line using ! before tile name", 
                    "# date_min the start date for this task, possible values:", 
                    "#    empty (2015-01-01 will be used) | any date | today-xx (xx nb of days before today to consider)", 
                    "# date_max the last date for this task, possible values:", 
                    "#    empty (9999-12-31 will be used) | any date | today", 
                    "# max_clouds: max cloud cover to consider for downloading images, and computing l2a products ", 
                    "# l1c: download l1c: True|False", 
                    "# l2a: compute l2a with sen2chain: True | False", 
                    "# cloudmasks: the cloudmask(s) to compute and use to mask indice(s). Possible values range from none (False) to multiple cloudmasks:", 
                    "#    False | CM001/CM002/CM003-PRB1-ITER5/CM004-CSH1-CMP1-CHP1-TCI1-ITER0/etc.", 
                    "# indices: False | All | NDVI/NDWIGAO/etc.", 
                    "# remove: used to remove downloaded L1C and/or produced L2A, possible values: False | l1c | l2a | l1c/l2a", 
                    "# comments: free user comments, ie tile name, etc.", 
                    "",
                    "",
                ]
            )
            header = "\n".join(
                [
                    "logs = " + str(self.logs),
                    "timing = " + self.timing,
                    "provider = " + self.provider,
                    "tries = " + str(self.tries),
                    "sleep = " + str(self.sleep),
                    "nb_proc = " + str(self.nb_proc),
                    "copy_l2a_sideproducts = " + str(self.copy_l2a_sideproducts),
                    "clean_before = " + str(self.clean_before),
                    "clean_after = " + str(self.clean_after),
                    "process_by_line = " + str(self.process_by_line),
                    "",
                    "",
                ]
            )
            
            for line in comments_header:
                ict.write(line)
            
            for line in header:
                ict.write(line)
            # self.tasks.to_csv(ict)
            self.unformat_cm_indices()
            self.tasks.to_csv(ict, index=False, sep=";")
            self.format_cm_indices()      


    def get_cron_status(self):
        """
        Returns job cron status.
        """
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                if job.is_enabled():
                    status = "enabled"
                else:
                    status = "disabled"
                self.cron_job = job
                self.cron_timing = str(self.cron_job.slices)
        else:
            status = "absent"
            self.cron_job = None
            self.cron_timing = None
        return status

    def create_python_script(self):
        """
        Creates python script for cron.
        """
        lines = ["# -*- coding:utf-8 -*-\n"]
        lines.append("\n")
        lines.append('"""\n')
        lines.append("Module for managing sen2chain processing jobs\n")
        lines.append('"""\n')
        lines.append("\n")
        lines.append("from sen2chain import Job\n")
        lines.append(
            'Job("'
            + self.jid
            + '").run(clean_before = False, clean_after = False)\n'
        )
        with open(str(self._python_script_path), "w") as f:
            f.writelines(lines)

    def cron_enable(
        self,
    ):
        """
        Enables job in cron.
        """
        self.save()
        self.create_python_script()
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Enabling job")
                if self.timing:
                    job.setall(self.timing)
                job.enable()
        else:
            job = self._cron.new(
                command="/usr/bin/python3 " + str(self._python_script_path),
                comment="sen2chain_job_" + self.jid,
            )
            if self.timing:
                job.setall(self.timing)
            else:
                job.setall("0 0 * * *")
            job.enable()
            logger.info("Enabling job...")

        # logger.info("Time: {}".format(job.time))
        self._cron.write()
        # new.enable(False)
        self.get_cron_status()

    def cron_disable(self):
        """
        Disable job in cron. The corresponding line is commented in cron.
        """
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Disabling job...")
                job.enable(False)
            self._cron.write()
            self.get_cron_status()

    def cron_remove(self):
        """
        Remove job from cron. The corresponding line is deleted
        """
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Removing job from cron...")
                self._cron.remove(job)
            self._cron.write()
            self.get_cron_status()

    def read(self, path):
        parser = ConfigParser(allow_no_value=True, strict = False)
        with open(str(path)) as stream:
            parser.read_string(
                "[top]\n" + stream.read(),
            )  # This line does the trick.
        self.logs = bool(
            setuptools.distutils.util.strtobool(
                parser["top"]["logs"]
            )
        )
        self.timing = parser["top"]["timing"]
        try:
            self.provider = parser["top"]["provider"]
        except:
            self.provider = "peps"
        try:
            self.tries = int(parser["top"]["tries"])
        except:
            self.tries = 1  
        try:
            self.sleep = int(parser["top"]["sleep"])
        except:
            self.sleep = 0
        try:
            self.nb_proc = int(parser["top"]["nb_proc"])
        except:
            self.nb_proc = 8
        try:
            self.copy_l2a_sideproducts = bool(
                setuptools.distutils.util.strtobool(
                    parser["top"]["copy_l2a_sideproducts"]
                )
            )
        except:
            self.copy_l2a_sideproducts = False
        try:
            self.clean_before = bool(setuptools.distutils.util.strtobool(
                    parser["top"]["clean_before"]
                )
            )
        except:
            self.clean_before = True
        try:
            self.clean_after = bool(setuptools.distutils.util.strtobool(
                    parser["top"]["clean_after"]
                )
            )
        except:
            self.clean_after= True
        try:
            self.process_by_line = int(parser["top"]["process_by_line"])
        except:
            self.process_by_line = 0         
            
        for i in range(1, 11):
            try:
                self.tasks = pd.read_csv(
                    path,
                    sep=";",
                    na_values="",
                    na_filter=False,
                    comment="#",
                    dtype=str,
                    header = i,
                    skip_blank_lines = True,
                )
                if 'comments' not in self.tasks:
                    self.tasks['comments'] = ""
                if 'cloudmask' in self.tasks:
                    self.tasks.rename(
                        columns={"cloudmask": "cloudmasks"}, 
                        inplace=True,
                    )
                self.format_cm_indices()      
                break
            except:
                pass
                
    def format_cm_indices(self):
        for index, row in self.tasks.iterrows():
            if not row.indices == "False":
                if row.indices == "All":
                    self.tasks.at[
                        index, "indices"
                    ] = IndicesCollection.available_indices
                else:
                    if not isinstance(self.tasks.at[index, "indices"] , list):
                        self.tasks.at[index, "indices"] = str(row.indices).split(
                            "/"
                        )
            if not row.cloudmasks == "False":
                if not isinstance(self.tasks.at[index, "cloudmasks"] , list):
                    self.tasks.at[index, "cloudmasks"] = str(row.cloudmasks).split(
                        "/"
                    )
    
    def unformat_cm_indices(self):
        for index, row in self.tasks.iterrows():
            if not((row.indices == "False") or (row.indices == "All")):
                # logger.info(row.indices)
                self.tasks.at[index, "indices"] = '/'.join(self.tasks.at[index, "indices"])   
            if not (row.cloudmasks == "False"):
                self.tasks.at[index, "cloudmasks"] = '/'.join(self.tasks.at[index, "cloudmasks"])   
    
    def fill_dates_clouds(
        self,
        tasks, 
    ):
        """
        Fill in dates and cloud cover values with correct values.
        """
        for index, row in tasks.iterrows():
            if not row.max_clouds:
                tasks.at[index, "max_clouds"] = 100
            else:
                tasks.at[index, "max_clouds"] = int(tasks.at[index, "max_clouds"])
            if not row.date_min:
                # self.tasks.at[index, "start_time"] = (datetime.datetime.now()-datetime.timedelta(days=delta_t)).strftime('%Y-%m-%d')
                tasks.at[index, "date_min"] = datetime.datetime.strptime(
                    "2015-01-01", "%Y-%m-%d"
                ).strftime("%Y-%m-%d")
            elif "today" in row.date_min:
                tasks.at[index, "date_min"] = (
                    datetime.datetime.now() - datetime.timedelta(
                        days = int(row.date_min.split("-")[1])
                    )
                ).strftime("%Y-%m-%d") 
            if not row.date_max:
                tasks.at[index, "date_max"] = datetime.datetime.strptime(
                    "9999-12-31", "%Y-%m-%d"
                ).strftime("%Y-%m-%d")
            elif row.date_max == "today":
                tasks.at[index, "date_max"] = (
                    datetime.datetime.now() + datetime.timedelta(days=1)
                ).strftime("%Y-%m-%d")   
        return tasks

    
    @staticmethod
    def get_cm_version(identifier) -> str:
        """
        Returns cloudmask version from a cloud mask identifier string.
        :param identifier: string pattern from which the cloud mask version is extracted.
        :type identifier: str
        """
        try:
            pat = re.compile(r"(?P<cm_version>CM00[1-2])")
            returned_val = pat.match(identifier).groupdict()
        except:
            try:
                pat = re.compile(
                    r"(?P<cm_version>CM003)"
                    + "-PRB(?P<probability>.*)"
                    + "-ITER(?P<iterations>.*)"
                )
                returned_val = pat.match(identifier).groupdict()
            except:
                try:
                    pat = re.compile(
                        r"(?P<cm_version>CM004)"
                        + "-CSH(?P<cld_shad>.*)"
                        + "-CMP(?P<cld_med_prob>.*)"
                        + "-CHP(?P<cld_hi_prob>.*)"
                        + "-TCI(?P<thin_cir>.*)"
                        + "-ITER(?P<iterations>.*)"
                    )
                    returned_val = pat.match(identifier).groupdict()
                except:
                    returned_val = {
                        "cm_version": "CM001",
                        "probability": 1,
                        "iterations": 5,
                        "cld_shad": True,
                        "cld_med_prob": True,
                        "cld_hi_prob": True, 
                        "thin_cir": True, 
                    }        
        

        if "probability" not in returned_val:
            returned_val["probability"] = 1
        if "iterations" not in returned_val:
            returned_val["iterations"] = 5
        if "cld_shad" not in returned_val:
            returned_val["cld_shad"] = True
        if "cld_med_prob" not in returned_val:
            returned_val["cld_med_prob"] = True
        if "cld_hi_prob" not in returned_val:
            returned_val["cld_hi_prob"] = True
        if "thin_cir" not in returned_val:
            returned_val["thin_cir"] = True
        
        returned_val["probability"] = int(returned_val["probability"])
        returned_val["iterations"] = int(returned_val["iterations"])
        returned_val["cld_shad"] = returned_val["cld_shad"] in ["1", 1, True, "True"] 
        returned_val["cld_med_prob"] = returned_val["cld_med_prob"] in ["1", 1, True, "True"]
        returned_val["cld_hi_prob"] = returned_val["cld_hi_prob"] in ["1", 1, True, "True"]
        returned_val["thin_cir"] = returned_val["thin_cir"] in ["1", 1, True, "True"]

        return returned_val

    def run(
        self,
        # nb_proc: int = self.nb_proc,
        clean_before: bool = False,
        clean_after: bool = False,
    ):
        """
        Runs job immediately. Job needs to be saved first.
        Tasks are performed line by line.
        """
        
        tasks = copy.deepcopy(self.tasks)
        tasks = self.fill_dates_clouds(tasks)
        
        if self.logs:
            self._log_folder_path.mkdir(exist_ok=True)
            self._log_file = self._log_folder_path / (
                "job_"
                + self.jid
                + "_run_"
                + datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
                + ".log"
            )
            f = open(str(self._log_file), "w")
            f.write("{}\nStart\n\n".format(datetime.datetime.now()))
            f.write(repr(self) + "\n\n")
            f.flush()
        
        # tasks = self.tasks
        tasks = tasks[tasks["tile"].str.contains("!") == False]
        logger.info("Running task(s):\n{}".format(tasks))
        logger.info("Processing by lines: {}".format(self.process_by_line))
        
        # list_df = [df[i:i+n] for i in range(0,len(df),n)]
        
        if not tasks.empty:
            ## Cleaning before
            if clean_before or self.clean_before:
                logger.info("Cleaning Tiles")
                if self.logs:
                    f.write("{} - Cleaning Tiles\n\n".format(datetime.datetime.now()))
                    f.flush()
                
                clean_list = []
                for index, row in tasks.iterrows():
                    clean_list.append(row.tile)
                lib = Library()
                pb_before = lib.clean(
                    clean_list, remove=True, remove_indice_tif=True
                )
                # lib.clean(clean_list, remove=False)
            if self.process_by_line > 0:
                split_tasks = np.array_split(
                    tasks, 
                    tasks.shape[0] // self.process_by_line + min(1, tasks.shape[0] % self.process_by_line)
                )
            else:
                split_tasks = [tasks]
            for chunk in split_tasks:
            
                ## L1C download - each tile sequential
                # download_list= []
                # logger.info("Downloading l1c seq")
                # for index, row in chunk.iterrows():
                    # if bool(setuptools.distutils.util.strtobool(str(row.l1c))):
                        # t = Tile(row.tile)
                        # logger.info("Tile: {}".format(t.name))
                        # downloaded = t.get_l1c(
                            # provider = "peps",
                            # download = True,
                            # dl_mode = "multit",
                            # start = row.date_min,
                            # end = row.date_max,
                            # new = False,
                        # )
                        # logger.info("Downloaded list: {}".format(downloaded))
                        # download_list.extend(downloaded)
                # logger.info("download_list: {}".format(download_list))

                ## L1C download - all tiles multi
                logger.info("Downloading l1c products")
                if self.logs:
                    f.write("{}\nDownloading l1c products\n".format(datetime.datetime.now()))
                    f.flush()
                
                all_downloaded_products = []
                for essai in range(self.tries):
                    logger.info("Download try #{}/{}".format(essai+1, self.tries))
                    if self.logs:
                        f.write("Download try #{}/{}\n".format(essai+1, self.tries))
                        f.flush()
                    ## Get new product list to download
                    download_list= []
                    for index, row in chunk.iterrows():
                        ref = "l1c"
                        if bool(setuptools.distutils.util.strtobool(str(row.l1c))):
                            if not (row.remove == "False" or not row.remove):
                                if "l1c" in str(row.remove).lower():
                                    if "l2a" not in str(row.remove).lower():
                                        ref = "l2a"
                                    else:
                                        ref = []
                                        for cm in row.cloudmasks:
                                            if not (row.cloudmasks == "False" or not row.cloudmasks):
                                                ref.append(cm)
                                            for ind in row.indices:
                                                if not (row.indices == "False" or not row.indices):
                                                    ref.extend([ind, ind + "_" + cm])
                                        ref = list(set(ref))
                                                
                            t = Tile(row.tile)
                            tile_download_list = t.get_l1c(
                                provider = self.provider,
                                download = False,
                                start = row.date_min,
                                end = row.date_max,
                                new = False,
                                ref = ref,
                                max_cloudcover = row.max_clouds,
                            )
                            download_list.extend(tile_download_list)
                            logger.info("Checked tile: {} - {} new product(s) to download".format(t.name, len(tile_download_list)))
                            if self.logs:
                                if len(tile_download_list):
                                    f.write("{}\n{} - {} new product(s) to download\n".format(
                                        datetime.datetime.now(),
                                        t.name, 
                                        len(tile_download_list)), 
                                    )
                                    f.flush()
                    logger.info("download_list: {}".format(download_list))
                    ### download 
                    before_list = [p.location for p in download_list]
                    if essai == 0:
                        total_before_list = [p.location for p in download_list]
                    logger.info("{} product(s) to download".format(len(download_list)))
                    if self.logs:
                        f.write("Total: {} product(s) to download\n".format(len(download_list)))
                        f.flush()
                    # S2cEodag_download(download_list, "seq")
                    S2cEodag_download(download_list, "multit")
                    # S2cEodag_download(download_list, "multip")
                    after_list = [p.location for p in download_list]
                    downloaded_products = [Path(after_list[i]).stem for i in range(len(before_list)) if before_list[i] != after_list[i]]
                    all_downloaded_products.extend(downloaded_products)
                    if self.logs:
                        f.write("{}\n".format(datetime.datetime.now()))
                        f.flush()       
                    if len(downloaded_products) == len(before_list):
                        logger.info("Downloaded all {} product(s)".format(len(downloaded_products), downloaded_products))
                        if downloaded_products:
                            logger.info("{}".format(downloaded_products))
                        if self.logs:
                            f.write("Downloaded all {} product(s)\n".format(len(downloaded_products)))
                            if downloaded_products:
                                for pro in downloaded_products:
                                    f.write("{}\n".format(pro))
                                f.write("\n")
                            f.flush()                    
                        break
                    else:
                        logger.info("Downloaded {}/{} product(s)".format(len(downloaded_products), len(before_list)))
                        if self.logs:
                            f.write("Downloaded {}/{} product(s)\n".format(len(downloaded_products), len(before_list)))
                            if downloaded_products:
                                for pro in downloaded_products:
                                    f.write("{}\n".format(pro))
                            if essai == self.tries-1:
                                f.write("\n")
                            f.flush()                    
                    if (self.tries > 1) and (essai < (self.tries-1)):
                        logger.info("Sleeping {}min before retrying".format(self.sleep))
                        if self.logs:
                            f.write("Sleeping {}min before retrying\n".format(self.sleep))
                            f.flush()
                        time.sleep(self.sleep*60)
                
                # remove duplicates from all_downloaded_products list
                all_downloaded_products.sort()
                uniq = list(k for k,_ in groupby(all_downloaded_products))
                all_downloaded_products = uniq
                
                logger.info(
                    "Total {}/{} downloaded product(s)".format(
                        len(all_downloaded_products),
                        len(total_before_list),
                    )
                )
                if self.logs:
                    f.write(
                        "Total {}/{} downloaded product(s)\n\n".format(
                            len(all_downloaded_products),
                            len(total_before_list),
                        )
                    )
                    f.flush()
                
                # Traitement des L1C en L2A
                logger.info("Computing l2a")
                if self.logs:
                    f.write("{}\nComputing l2a\n".format(datetime.datetime.now()))
                    f.flush()

                l1c_process_list = []
                for index, row in chunk.iterrows():
                    if bool(setuptools.distutils.util.strtobool(str(row.l2a))):
                        t = Tile(row.tile)
                        l1c_to_process = list(
                            [
                                p.identifier,
                                self.copy_l2a_sideproducts,
                                self.remove_used_l1c,
                            ]
                            for p in t.l2a_missings.filter_dates(
                                date_min=row.date_min, date_max=row.date_max
                            ).filter_clouds(cover_max = row.max_clouds)
                        )
                        if len(l1c_to_process):
                            logger.info(
                                "{} adding {} l1c files".format(
                                    row.tile, len(l1c_to_process)
                                )
                            )
                        l1c_process_list.extend(l1c_to_process)
                        
                # l1c_process_list = list(
                    # set(
                        # chain.from_iterable(l1c_process_list)
                    # )
                # )
                
                ### remove duplicate l1c from l1c_process_list
                l1c_process_list.sort()
                uniq = list(k for k,_ in groupby(l1c_process_list))
                l1c_process_list = uniq
                
                logger.info(
                    "Process list ({} files): {}".format(
                        len(l1c_process_list), l1c_process_list
                    )
                )
                if self.logs:
                    f.write(
                        "Process list ({} files): {}\n\n".format(
                            len(l1c_process_list), l1c_process_list
                        )
                    )
                    f.flush()
                
                ### Computing L2A
                l2a_res = False
                if l1c_process_list:
                    l2a_res = l2a_multiprocessing(
                        l1c_process_list, nb_proc = self.nb_proc
                    )

                # Remove downloaded L1C
                for index, row in chunk.iterrows():
                    if "l1c" in str(row.remove).lower():
                        t = Tile(row.tile)
                        prodlist = [p for p in all_downloaded_products if row.tile in p]
                        prodlist_cleaned = [
                            p for p in prodlist if p in [Path(q.identifier).stem for q in t.l1c]
                        ]
                        prodlist_skipped = [
                            p for p in prodlist if p not in [Path(q.identifier).stem for q in t.l1c]
                        ]
                        logger.info(
                            "{} removing downloaded l1c product(s): {}".format(
                                row.tile,
                                prodlist_cleaned,
                            )
                        )
                        if prodlist_skipped:
                            logger.info(
                                "{} skipping {} l1c product(s), download error?:\n{}".format(
                                    row.tile,
                                    len(prodlist_skipped),
                                    prodlist_skipped,
                                )
                            )                    
                        if self.logs:
                            f.write(
                                "{}\nRemoving downloaded l1c products for tile: {}\n{}\n".format(
                                    datetime.datetime.now(), 
                                    row.tile,
                                    prodlist_cleaned,
                                )
                            )
                            if prodlist_skipped:
                                f.write(
                                    "\nSkipping {} l1c product(s), download error?:\n{}".format(
                                        len(prodlist_skipped),
                                        prodlist_skipped,
                                    )
                                )  
                            else:
                                f.write("\n")
                            f.flush()                      
                        t.remove_l1c(prodlist)

                # Comuting cloudmasks (from L2A)
                logger.info("Computing cloudmasks")
                if self.logs:
                    f.write("{}\nComputing cloudmasks\n".format(datetime.datetime.now()))
                    f.flush()
                reprocess = False
                cld_l2a_process_list = []
                for index, row in chunk.iterrows():
                    # if not bool(distutils.util.strtobool(str(row.cloudmask))):
                    if not (row.cloudmasks == "False" or not row.cloudmasks):
                        for cm in row.cloudmasks:
                            cloudmask = self.get_cm_version(cm)                    
                            t = Tile(row.tile)
                            l2a_to_process = [
                                p.identifier
                                for p in t.cloudmasks_missing(
                                    cm_version = cloudmask["cm_version"].upper(),
                                    probability = cloudmask["probability"],
                                    iterations = cloudmask["iterations"],
                                    cld_shad = cloudmask["cld_shad"],
                                    cld_med_prob = cloudmask["cld_med_prob"],
                                    cld_hi_prob = cloudmask["cld_hi_prob"],
                                    thin_cir = cloudmask["thin_cir"],                       
                                ).filter_dates(
                                    date_min=row.date_min, date_max=row.date_max
                                )
                            ]
                            for j in l2a_to_process:
                                l2a_cm_details = [
                                    j,
                                    cloudmask["cm_version"].upper(),
                                    cloudmask["probability"],
                                    cloudmask["iterations"],
                                    cloudmask["cld_shad"],
                                    cloudmask["cld_med_prob"],
                                    cloudmask["cld_hi_prob"],
                                    cloudmask["thin_cir"],
                                    reprocess,
                                ]
                                cld_l2a_process_list.append(l2a_cm_details)
                            if len(l2a_to_process):
                                logger.info(
                                    "{} - {} - {} l2a product(s) to process".format(
                                        row.tile, 
                                        cm,
                                        len(l2a_to_process),
                                    )
                                )
                                if self.logs:
                                    if len(l2a_to_process):
                                        f.write(
                                            "{} - {} - {} l2a product(s) to process\n".format(
                                                row.tile, 
                                                cm,
                                                len(l2a_to_process),
                                            )
                                        )
                                        f.flush()
                # removing duplicate cloudmasks before computing
                cld_l2a_process_list.sort()
                uniq = list(k for k,_ in groupby(cld_l2a_process_list))
                cld_l2a_process_list = uniq
                
                logger.info(
                    "Cloudmask computing process list ({} l2a products): {}".format(
                        len(cld_l2a_process_list), cld_l2a_process_list
                    )
                )
                if self.logs:
                    f.write(
                        "Cloudmask computing process list ({} l2a products):\n".format(
                            len(cld_l2a_process_list),                     )
                    )
                    for cld in cld_l2a_process_list:
                        f.write("{}\n".format(cld))
                    f.write("\n")
                    f.flush()
                    
                cld_res = False
                if cld_l2a_process_list:
                    cld_res = cld_version_probability_iterations_reprocessing_multiprocessing(
                        cld_l2a_process_list, nb_proc = self.nb_proc
                    )
                
                # Traitement des L2A (indices)
                logger.info("Computing indices")
                if self.logs:
                    f.write("{}\nComputing indices\n".format(datetime.datetime.now()))
                    f.flush()
                
                quicklook = False
                indices_l2a_process_list = []
                for index, row in chunk.iterrows():
                    if not (row.indices == "False" or not row.indices):
                        nodata_clouds = not (row.cloudmasks == "False" or not row.cloudmasks)
                        if nodata_clouds:
                            for cm in row.cloudmasks:
                                cloudmask = self.get_cm_version(cm) 
                                t = Tile(row.tile)
                                for i in row.indices:
                                    l2a_to_process = [
                                        p.identifier
                                        for p in t.missing_indices(
                                            indice=i,
                                            nodata_clouds=nodata_clouds,
                                            cm_version = cloudmask["cm_version"].upper(),
                                            probability = cloudmask["probability"],
                                            iterations = cloudmask["iterations"],
                                            cld_shad = cloudmask["cld_shad"],
                                            cld_med_prob = cloudmask["cld_med_prob"],
                                            cld_hi_prob = cloudmask["cld_hi_prob"],
                                            thin_cir = cloudmask["thin_cir"],                                  
                                        ).filter_dates(
                                            date_min=row.date_min, date_max=row.date_max
                                        )
                                    ]
                                    for j in l2a_to_process:
                                        l2a_ind_details = [
                                            j,
                                            i,
                                            reprocess,
                                            nodata_clouds,
                                            quicklook,
                                            cloudmask["cm_version"].upper(),
                                            cloudmask["probability"],
                                            cloudmask["iterations"],
                                            cloudmask["cld_shad"],
                                            cloudmask["cld_med_prob"],
                                            cloudmask["cld_hi_prob"],
                                            cloudmask["thin_cir"],
                                        ]
                                        indices_l2a_process_list.append(l2a_ind_details)
                                    if len(l2a_to_process):
                                        logger.info(
                                            "{} - {} - {} - {} l2a product(s) to process".format(
                                                row.tile, 
                                                cm,
                                                i, 
                                                len(l2a_to_process)
                                            )
                                        )
                                        if self.logs:
                                            if len(l2a_to_process):
                                                f.write(
                                                    "{} - {} - {} - {} l2a product(s) to process\n".format(
                                                        row.tile, 
                                                        cm, 
                                                        i,
                                                        len(l2a_to_process),
                                                    )
                                                )
                                                f.flush()
                                    
                # removing duplicate indice computing
                indices_l2a_process_list.sort()
                uniq = list(k for k,_ in groupby(indices_l2a_process_list))
                indices_l2a_process_list = uniq
                
                                    
                logger.info(
                    "Indice computing process list ({} l2a products): {}".format(
                        len(indices_l2a_process_list), indices_l2a_process_list
                    )
                )
                if self.logs:
                    f.write(
                        "Indice computing process list ({} l2a products): \n".format(
                            len(indices_l2a_process_list),
                        )
                    )
                    for ind in indices_l2a_process_list:
                        f.write("{}\n".format(ind))
                    f.write("\n")
                    f.flush()            
                
                indices_res = False
                if indices_l2a_process_list:
                    indices_res = idx_multiprocessing(
                        indices_l2a_process_list, nb_proc = self.nb_proc
                    )

                # Remove L2A
                l2a_remove_list = [
                    l1c[0].replace(
                        "L1C_", "L2A_"
                    ).replace(
                        "_OPER_", "_USER_"
                    ).replace(
                        ".SAFE", ""
                    ) for l1c in l1c_process_list
                ]
                for index, row in chunk.iterrows():
                    if "l2a" in str(row.remove).lower():
                        t = Tile(row.tile)
                        prodlist = [p for p in l2a_remove_list if row.tile in p]
                        prodlist_cleaned = [
                            p for p in prodlist if p in [Path(q.identifier).stem for q in t.l2a]
                        ]
                        prodlist_skipped = [
                            p for p in prodlist if p not in [Path(q.identifier).stem for q in t.l2a]
                        ]
                        logger.info(
                            "{} removing produced l2a product(s): {}".format(
                                row.tile,
                                prodlist_cleaned,
                            )
                        )
                        if prodlist_skipped:
                            logger.info(
                                "{} skipping {} l2a product(s), Sen2Cor error?:\n{}".format(
                                    row.tile,
                                    len(prodlist_skipped),
                                    prodlist_skipped,
                                )
                            )                          
                        if self.logs:
                            f.write(
                                "{}\nRemoving produced l2a products for tile: {}\n{}\n".format(
                                    datetime.datetime.now(), 
                                    row.tile,
                                    prodlist_cleaned,
                                )
                            )
                            if prodlist_skipped:
                                f.write(
                                    "\nSkipping {} l2a product(s), download error?:\n{}".format(
                                        len(prodlist_skipped),
                                        prodlist_skipped,
                                    )
                                )  
                            else:
                                f.write("\n")
                            f.flush()                      
                        t.remove_l2a(prodlist)

            # Cleaning after
            if clean_after or self.clean_after:
                logger.info("Cleaning Tiles")
                if self.logs:
                    f.write("{}\nCleaning Tiles\n\n".format(datetime.datetime.now()))
                    f.flush()
                
                clean_list = []
                for index, row in tasks.iterrows():
                    clean_list.append(row.tile)
                lib = Library()
                pb_before = lib.clean(
                    clean_list, remove=True, remove_indice_tif=True
                )
                # lib.clean(clean_list, remove=False)

        else:
            logger.info("No task defined for this job, doing nothing")
            if self.logs:
                f.write("No task defined for this job, doing nothing\n\n")
                f.flush()

        if self.logs:
            f.write("{}\nFin".format(datetime.datetime.now()))
            f.close()
